from typing import Final

LIBRARY_NAME_SINGING_TOOL_MODEL: Final[str] = (
    "SingingTool.Model, Version=1.0.0.0, Culture=neutral, PublicKeyToken=null"
)
LIBRARY_NAME_SINGING_TOOL_LIBRARY: Final[str] = (
    "SingingTool.Library, Version=1.0.0.0, Culture=neutral, PublicKeyToken=null"
)

# Values
VALUE_LIST_VERSION_SONG_TEMPO: Final[int] = 2
VALUE_LIST_VERSION_SONG_BEAT: Final[int] = 2
VALUE_LIST_VERSION_SONG_ITRACK: Final[int] = 4
VALUE_LIST_VERSION_SONG_NOTE: Final[int] = 1054
