CS_DBLCLKS = 8
GWL_STYLE = -16
GWL_EXSTYLE = -20
HTCAPTION = 2
HTSYSMENU = 3
HTMINBUTTON = 8
HTMAXBUTTON = 9
HTLEFT = 10
HTRIGHT = 11
HTTOP = 12
HTTOPLEFT = 13
HTTOPRIGHT = 14
HTBOTTOM = 15
HTBOTTOMLEFT = 16
HTBOTTOMRIGHT = 17
HTBORDER = 18
HTCLOSE = 20
SC_MAXIMIZE = 61488
SC_RESTORE = 61728
SIZE_MINIMIZED = 1
TPM_LEFTBUTTON = 0
TPM_RIGHTBUTTON = 2
TPM_LEFTALIGN = 0
TPM_CENTERALIGN = 4
TPM_RIGHTALIGN = 8
TPM_TOPALIGN = 0
TPM_VCENTERALIGN = 16
TPM_BOTTOMALIGN = 32
TPM_HORIZONTAL = 0
TPM_VERTICAL = 64
TPM_NONOTIFY = 128
TPM_RETURNCMD = 256
TPM_RECURSE = 1
WS_EX_LAYERED = 0x00080000
WS_MAXIMIZEBOX = 65536
WS_CAPTION = 12582912
WS_THICKFRAME = 262144
WM_NCCALCSIZE = 131
WM_NCHITTEST = 132
WM_NCMOUSEMOVE = 160
WM_NCLBUTTONDOWN = 161
WM_NCLBUTTONUP = 162
WM_NCLBUTTONDBLCLK = 163
WM_NCMOUSEHOVER = 672
WM_NCMOUSELEAVE = 674
WM_SIZE = 5
WM_SYSCOMMAND = 274
WM_ACTIVATE = 6
WVR_HREDRAW = 256
WVR_VREDRAW = 512
WVR_REDRAW = WVR_HREDRAW | WVR_VREDRAW
