from typing import Final

DEFAULT_PHONEME: Final[str] = "r,a"
TIME_UNIT_AS_TICKS_PER_BPM: Final[float] = 4.8 / 120
OCTAVE_OFFSET: Final[int] = -1
TICK_RATE: Final[float] = 2.0
MIN_DATA_LENGTH: Final[int] = 500
TEMP_VALUE_AS_NULL: Final[float] = -1.0
