from typing import Final

DEFAULT_CHINESE_PHONEME: Final[str] = "l a"
DEFAULT_JAPANESE_PHONEME: Final[str] = "4 a"
PITCH_BEND_NAME: Final[str] = "pitchBend"
PITCH_BEND_SENSITIVITY_NAME: Final[str] = "pitchBendSens"
BPM_RATE: Final[float] = 100.0
